-- stuff we use
local pairs, ipairs, tostring = pairs, ipairs, tostring

local Font                   = require("jive.ui.Font")
local package = package
local oo                     = require("loop.simple")

local string                 = require("string")
local table                  = require("jive.utils.table")
local io                     = require("io")

local Applet                 = require("jive.Applet")
local Framework              = require("jive.ui.Framework")
local Label                  = require("jive.ui.Label")
local SimpleMenu             = require("jive.ui.SimpleMenu")
local Textarea               = require("jive.ui.Textarea")
local Window                 = require("jive.ui.Window")
local Popup                  = require("jive.ui.Popup")
local Choice                 = require("jive.ui.Choice")

local socket           = require("socket")
local SocketHttp       = require("jive.net.SocketHttp")
local RequestHttp      = require("jive.net.RequestHttp")
local appletManager    = appletManager
local RequestJsonRpc = require("jive.net.RequestJsonRpc")
local Tile                   = require("jive.ui.Tile")

local MyTextArea = require("applets.Lyrics.MyTextArea")

local EVENT_KEY_PRESS  = jive.ui.EVENT_KEY_PRESS
local EVENT_UNUSED     = jive.ui.EVENT_UNUSED
local EVENT_SCROLL      = jive.ui.EVENT_SCROLL 


local jnt              = jnt

local log                    = require("jive.utils.log").logger("applets.setup")

module(...)
oo.class(_M, Applet)


local URLTEMPLATE_LIST = "/lookup.php?q=%s-%s&for=fullt&agent=slim"
local URLTEMPLATE_GET = "/getlyr.php?q=%s"
local URLTEMPLATE_KLIST = "/karlookup.php?q=%s-%s"
local URLTEMPLATE_KGET = "/showkar.php?q=%s&useid=1"
local LYRDB_HOST = "webservices.lyrdb.com"

-- Straight from the book
function cgiEscape(s)
   s = s:gsub("[&=+%%%c]", function (c)
			      return string.format("%%%02X", string.byte(c))
			   end)
   s = s:gsub(" ", "+")
   return s
end
   
function HandleLyricsSingle(self, r, lyricId, nk)
   log:debug("HandleLyricsSingle: " .. r)
   local window = Window("window", string.format("%s - %s",  lyricId[2], lyricId[3]))
   -- Eliminate \r chars and blank lines
   r = r:gsub("\r+", "")
   r = r:gsub("\n+", "\n")
   r = r:gsub("^\n%s*$", "")
   r = r:gsub("\t", "")
   mta = MyTextArea.MyTextArea(r, nk, window, function(syncer) syncToCurrentTime(self, syncer) end)
   mta:draw()
   mta:setEventHandler()
   self:tieAndShowWindow(window)
   return window
end
	 

function getSingleLyric(self, lyricId, nk)
   log:debug(string.format("getSingleLyric: %s/%s/%s", lyricId[1], lyricId[2], lyricId[3]))
   local url = string.format(URLTEMPLATE_KGET, lyricId[1])
   if lyricId[4] then
      url = string.format(URLTEMPLATE_GET, lyricId[1])
   end
      
   local ip, err = socket.dns.toip(LYRDB_HOST)
   self.http = SocketHttp(jnt, ip, 80)
   local req = RequestHttp(function(chunk, error)
			      if error then
				 log:error("Error getting lyric for " .. lyricId[1] .. " " .. error)
			      elseif chunk then
				 HandleLyricsSingle(self, chunk, lyricId, nk)
			      else
				 log:error("No lyric data returned")
			      end
			   end,
			   'GET',
			   url,
			   {headers = {Host = LYRDB_HOST}}
			)
   self.http:fetch(req)
   
end


function HandleLyricsList(self, r, nk)
   log:debug("HandleLyricsList: " .. r)
   local window = Window("window", "Lyric Matches")
   window:addListener(EVENT_KEY_PRESS, function(event) return EVENT_UNUSED end)
   
   if r:match("^error:.*") then
      Popup("popup", string.format("Server %s", r)):show()
      return 0
   end

   -- Parse into a table of lines, each line is a table of words
   -- words[1] = ref words[2] = Artist words[3] = Title words[4] = Karaoke/Non-Karaoke
   r = r.."\n"
   for l in r:gmatch("(.-)\n") do
      -- Append a \\ to make split easier
      l = l.."\\"
      words = {}
      if l:match('\\.*\\') then
	 for w in l:gmatch("(.-)\\(.-)") do
	    words[#words + 1] = w
	 end
	 words[#words + 1] = nk
	 self.llist[#self.llist + 1] = words
      end
      -- Need 5 karaoke, 5 non-karaoke
      if #self.llist % 5 == 0 then
	 break
      end
   end

   -- Display artist + title menu items, but only if karaoke + non-karaoke items complete
   if nk then
      local menu = SimpleMenu("menu", "gaga")
      window:addWidget(menu)
      local items = {}
      for i, v in ipairs(self.llist) do
	 local ks = "k"
	 if v[4] then
	    ks = "n"
	 end
	 items[#items + 1] = {text = string.format("%s %s - %s", ks, v[2], v[3]),
			      callback = function(event, menuItem) getSingleLyric(self, v) end
			   }
      end
      menu:setItems(items)
      self:tieAndShowWindow(window)
   end
   return #self.llist
end


function myCli(self, playerid, server, cmd, fn)
   local req = RequestJsonRpc(function(chunk, err)
				 if err then
				    log:error("Json error: " .. err)
				 else
				    fn(self, playerid, server, chunk)
				 end
			      end,
			      '/jsonrpc.js',
			      'slim.request', 
			      {playerid, {cmd, "?"}},
			      options)
   http = SocketHttp(jnt, server, 9000)
   http:fetch(req)
end



function getLyricsFor(self, artist, title, nk)
   if artist and title then
	  log:debug(string.format("getLyricsFor artist: '%s' title '%s'", artist, title))
      self.window:addWidget(Textarea("textarea", string.format("Lyrics for %s - %s", artist, title)))
      -- Try Karaoke first
      local ip, err = socket.dns.toip(LYRDB_HOST)
      self.http = SocketHttp(jnt, ip, 80)
      local urlt = URLTEMPLATE_KLIST
      if nk then
	 urlt = URLTEMPLATE_LIST
      end
      local url = string.format(urlt, cgiEscape(artist), cgiEscape(title))
      local req = RequestHttp(function(chunk, error)
				 if error then
				    log:error("lyrdb list request error: " .. error)
				 end
				 if chunk then
				    log:debug("Karaoke lookup")
				    local nmatches = HandleLyricsList(self, chunk, nk == "nk")
				    if nmatches <  10 then
				       if nk then
					  log:error("No lyrics match")
				       else
					  log:debug("Normal lookup")
					  self:getLyricsFor(artist, title, "nk")
				       end
				    end
				 end
			      end,
			      'GET',
			      url,
			      {headers = {Host = LYRDB_HOST}}
			   )
      self.http:fetch(req)
   else
      -- No artist and title. Meta data screwed?
      Popup("popup", string.format("No artist (%s) AND title (%s)", artist or '', title or '')):show()
   end
end

function GetCurrentlyPlaying(self)
   local discovery = appletManager:getAppletInstance("SlimDiscovery")
   local thePlayer = discovery:getCurrentPlayer()
   if thePlayer then
      local slimServerIP = thePlayer:getSlimServer():getIpPort()
      local playerid = thePlayer:getId()
      log:debug("player: " .. playerid .. " server: " .. slimServerIP)
      
      self:myCli(playerid, slimServerIP, "remote",
		 function(self, playerid, server, chunk)
		    log:debug('Remote: ' .. chunk.result._remote)
		    if chunk.result._remote == 0 then
		       -- Local
		       self:myCli(playerid, slimServerIP, "artist",
				  function(self, playerid, server, chunk)
				     local artist = chunk.result._artist
				     self:myCli(playerid, slimServerIP, "title",
						function(self, playerid, server, chunk)
						   local title = chunk.result._title
						   self:getLyricsFor(artist, title)
						end)
				  end)
		    else
		       -- Wiimote, works for RadioParadise
		       self:myCli(playerid, slimServerIP, "current_title",
				  function(self, playerid, server, chunk)
				     local artist, title = chunk.result._current_title:gmatch("(.*)%s*-%s*(.*)")()
				     self:getLyricsFor(artist, title)
				  end)
		       
		    end
		 end)
   else
      Popup("popup", "No current player"):show()
   end
end

function syncToCurrentTime(self, syncer)
   -- This is silly, we alread know these
   local discovery = appletManager:getAppletInstance("SlimDiscovery")
   local thePlayer = discovery:getCurrentPlayer()
   local slimServerIP = thePlayer:getSlimServer():getIpPort()
   local playerid = thePlayer:getId()
   self:myCli(playerid, slimServerIP, "remote",
	      function(self, playerid, server, chunk)
		 if chunk.result._remote == 0 then
		    -- Local only
		    self:myCli(playerid, slimServerIP, "time",
			       function(self, playerid, server, chunk)
				  syncer(chunk.result._time)
			       end)
		    
		 else
		    -- Remote time is elapsed since tuned in, not the start of a track, so sync to 1st line
		    syncer(0, true)
		 end
	      end)
end
   

function LyricsAppletShow(self)
   self.window = Window("window", "Lyrics Display")
   self.llist = {}
   self:GetCurrentlyPlaying()
   return self.window
end


function skin(self, s) 
   s.lyrics = {}
   s.lyrics.w, h = Framework:getScreenSize()
   s.lyrics.padding = { 10, 0, 0, 10 }
   s.lyrics.font = Font:load("fonts/" .. "FreeSansBold.ttf", 14)
   s.lyrics.fg = { 0xE7, 0xE7, 0xE7 }
   s.lyrics.sh = { 0x37, 0x37, 0x37 }
   s.lyrics.align = "left"

   s.current = {}
   s.current.padding = { 10, 0, 0, 10 }
   s.current.font = Font:load("fonts/" .. "FreeSansBold.ttf", 14)
   s.current.fg = { 0x28, 0xe3, 0x17 }
   s.current.align = "left"
end
